import os
import re
from re import Pattern
from typing import Callable, List

from gitignore_parser import parse_gitignore  # type: ignore
from gitignore_parser import IgnoreRule, handle_negation, rule_from_pattern

from zlipy.domain.filesfilter.constants import GITIGNORE_FILENAME, FilesFilterTypes
from zlipy.domain.filesfilter.interfaces import IFilesFilter


class GitIgnoreFilesFilter(IFilesFilter):
    def __init__(self, filename):
        self.filename = filename
        self._matches_func = self._load__matches_func()

    def _load__matches_func(self):
        try:
            matches = parse_gitignore(".gitignore")
        except Exception:
            matches = lambda x: False

        return matches

    def ignore(self, relative_path: str) -> bool:
        return self._matches_func(relative_path)


class AllowedExtensionsFilesFilter(IFilesFilter):
    def __init__(self) -> None:
        super().__init__()

        # fmt: off
        self._allowed_extensions = {
             ".py",   # Python files
             ".txt",  # Text files
             ".md",   # Markdown files
             ".json", # JSON files
             ".csv",  # Comma-separated values
             ".xml",  # XML files
             ".html", # HTML files
             ".css",  # CSS files
             ".ini",  # INI configuration files
             ".yaml", ".yml",  # YAML files
             ".java", # Java source files
             ".js",   # JavaScript files
             ".c",    # C source files
             ".cpp",  # C++ source files
             ".h",    # Header files
             ".hpp",  # C++ header files
             ".rb",   # Ruby files
             ".php",  # PHP files
             ".go",   # Go source files
             ".rs",   # Rust source files
             ".kt",   # Kotlin source files
             ".sh",   # Shell script files
             ".sql",  # SQL script files
             ".log",  # Log files
             ".env",  # Environment files
        }
        # fmt: on

    def ignore(self, relative_path: str) -> bool:
        _, extenstion = os.path.splitext(relative_path)
        return extenstion not in self._allowed_extensions


class MergeFilesFilter(IFilesFilter):
    def __init__(self, *args: IFilesFilter) -> None:
        super().__init__()

        self._filters = args

    def ignore(self, relative_path: str) -> bool:
        return any(filter.ignore(relative_path) for filter in self._filters)


class IgnoredFilesFilter(IFilesFilter):
    def __init__(self, patterns: List[str]):
        """Initialize the filter with a list of .gitignore-like patterns."""
        self.patterns = patterns
        # self.rules: List[IgnoreRule] = self._convert_patterns_to_regex(patterns)
        self.matches = self._convert_patterns_to_match_func(patterns)

    def _convert_patterns_to_match_func(
        self, patterns: List[str]
    ) -> Callable[[str], bool]:
        rules = []
        for pattern in patterns:
            rule = rule_from_pattern(pattern, base_path=".")
            rules.append(rule)

        if not any(r.negation for r in rules):
            return lambda file_path: any(r.match(file_path) for r in rules)
        else:
            # We have negation rules. We can't use a simple "any" to evaluate them.
            # Later rules override earlier rules.
            return lambda file_path: handle_negation(file_path, rules)

    def ignore(self, relative_path: str) -> bool:
        """Check if the relative path matches any of the ignored patterns."""
        return self.matches(relative_path)



import abc


class IFilesFilter(abc.ABC):
    @abc.abstractmethod
    def ignore(self, relative_path: str) -> bool:
        pass


class IProjectStructureLoader(abc.ABC):
    @abc.abstractmethod
    def load(self) -> list[str | dict]:
        pass



import warnings

warnings.filterwarnings("ignore")

import click

from zlipy.api_client import run
from zlipy.config import init_config
from zlipy.config.factory import ConfigFactory


@click.group(invoke_without_command=True)
@click.pass_context
def main(ctx):
    if not ctx.invoked_subcommand:
        run(
            config=ConfigFactory.create(
                debug=False,
                boost=False,
                disable_markdown_formatting=False,
            )
        )


@main.command()
def init():
    """Initialize the configuration."""
    init_config()
    click.echo("Configuration initialized.")


@main.command()
@click.option(
    "--disable-markdown-formatting",
    "-dmf",
    is_flag=True,
    help="Disable markdown formatting in the console.",
    default=False,
)
@click.option(
    "--debug",
    "-d",
    is_flag=True,
    help="Enable debug mode (more verbose output).",
    default=False,
)
@click.option(
    "--deep-dive",
    "-dd",
    is_flag=True,
    help="Enable deep dive mode (advanced analysis).",
    default=False,
)
def chat(disable_markdown_formatting: bool, debug: bool, deep_dive: bool):
    """Start a chat."""
    run(
        config=ConfigFactory.create(
            debug=debug,
            deep_dive=deep_dive,
            disable_markdown_formatting=disable_markdown_formatting,
        )
    )


cli = main


if __name__ == "__main__":
    cli()



import os

from zlipy.domain.tools.interfaces import ITool


class LoadFileTool(ITool):
    async def run(self, input: str) -> str | None:
        relative_path = input
        root_dir = os.path.abspath(os.getcwd())

        full_path = os.path.join(
            root_dir,
            relative_path[1:] if relative_path.startswith("/") else relative_path,
        )

        if not os.path.exists(full_path):
            return None

        with open(full_path, "r") as file:
            return file.read()



import contextlib
import os

from langchain_community.document_loaders import TextLoader
from langchain_community.vectorstores import DeepLake  # type: ignore
from langchain_core.embeddings import Embeddings

from zlipy.config.interfaces import IConfig
from zlipy.domain.filesfilter import FilesFilterFactory, IFilesFilter
from zlipy.domain.tools.interfaces import ITool
from zlipy.services.embeddings import APIEmbeddings


def load_docs(config: IConfig) -> list:
    root_dir = os.getcwd()
    docs = []

    files_filter: IFilesFilter = FilesFilterFactory.create(
        ignore_patterns=config.ignored_patterns
    )

    for dirpath, _, filenames in os.walk(root_dir):
        for file in filenames:
            if not files_filter.ignore(os.path.join(dirpath, file)):
                with contextlib.suppress(Exception):
                    if config.debug:
                        print(f"Loading file: {os.path.join(dirpath, file)}")

                    loader = TextLoader(os.path.join(dirpath, file), encoding="utf-8")
                    temp_docs: list = loader.load_and_split()
                    for doc in temp_docs:
                        doc.metadata["path"] = os.path.join(dirpath, file).replace(
                            root_dir, ""
                        )
                    docs.extend(temp_docs)

    return docs


def get_db_retriever(config: IConfig):
    texts = load_docs(config=config)
    db = DeepLake.from_documents(texts, APIEmbeddings(config=config), overwrite=True)

    retriever = db.as_retriever()
    retriever.search_kwargs["distance_metric"] = "cos"
    retriever.search_kwargs["fetch_k"] = 6
    retriever.search_kwargs["k"] = 6

    return db, retriever


class CodeBaseSearch(ITool):
    def __init__(self, config: IConfig) -> None:
        self.db, self.retriever = get_db_retriever(config=config)

    async def run(self, input: str) -> list[dict]:
        docs = self.retriever.invoke(input)

        return [
            {
                "path": item.metadata["path"],
                "content": item.page_content,
            }
            for item in docs
        ]



from zlipy.config.interfaces import IConfig
from zlipy.domain.tools import CodeBaseSearch, ITool, LoadFileTool
from zlipy.services.api import APIClientFactory
from zlipy.services.client.clients import Client
from zlipy.services.client.interfaces import IClient


class ClientFactory:
    @staticmethod
    def create(config: IConfig) -> IClient:
        tools: dict[str, ITool] = {
            "search": CodeBaseSearch(config=config),
            "load": LoadFileTool(),
        }

        return Client(APIClientFactory.create(), config=config, tools=tools)



import fcntl
import os
import sys


class UnblockTTY:
    def __enter__(self):
        self.fd = sys.stdin.fileno()
        self.flags_save = fcntl.fcntl(self.fd, fcntl.F_GETFL)
        flags = self.flags_save & ~os.O_NONBLOCK
        fcntl.fcntl(self.fd, fcntl.F_SETFL, flags)

    def __exit__(self, *args):
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.flags_save)



import httpx
import websockets

from zlipy.services.api.constants import API_BASE
from zlipy.services.api.interfaces import IAPIClient


class APIClient(IAPIClient):
    def __init__(self, base: str = API_BASE):
        super().__init__()
        self.base = base
        self.ws_base = base.replace("https", "wss").replace("http", "ws")

    async def generate_embeddings(
        self, api_key: str, inputs: list[str]
    ) -> list[list[float]]:
        endpoint = f"{self.base}/tools/embeddings/?token={api_key}"

        async with httpx.AsyncClient() as client:
            response = await client.post(endpoint, json={"texts": inputs}, timeout=60)
            response.raise_for_status()

            return response.json()

    def connect(self, api_key: str) -> websockets.connect:
        endpoint = f"{self.ws_base}/ws/?token={api_key}"
        return websockets.connect(endpoint, ping_timeout=500, ping_interval=10)



from langchain_core.embeddings import Embeddings

from zlipy.config.interfaces import IConfig
from zlipy.services.api.factory import APIClientFactory
from zlipy.services.api.interfaces import IAPIClient


class APIEmbeddings(Embeddings):
    def __init__(self, config: IConfig) -> None:
        super().__init__()

        self.client: IAPIClient = APIClientFactory.create()
        self.config: IConfig = config

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        import asyncio

        return asyncio.run(self.aembed_documents(texts))

    def embed_query(self, text: str) -> list[float]:
        import asyncio

        return asyncio.run(self.aembed_query(text))

    async def aembed_documents(self, texts: list[str]) -> list[list[float]]:
        return await self.client.generate_embeddings(
            api_key=self.config.api_key, inputs=texts
        )

    async def aembed_query(self, text: str) -> list[float]:
        return (await self.client.generate_embeddings(self.config.api_key, [text]))[0]



import abc


class IConfig(abc.ABC):
    @property
    @abc.abstractmethod
    def api_key(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def debug(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def disable_markdown_formatting(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def ignored_patterns(self) -> list[str]:
        pass

    @property
    @abc.abstractmethod
    def deep_dive(self) -> bool:
        pass



from zlipy.domain.events.interfaces import IEvent


class Event(IEvent):
    def __init__(self, name: str, data: dict):
        self._name = name
        self._data = data

    @property
    def name(self) -> str:
        return self._name

    @property
    def data(self) -> dict:
        return self._data



import abc
from typing import Coroutine

import websockets


class IAPIClient(abc.ABC):
    @abc.abstractmethod
    async def generate_embeddings(
        self, api_key: str, inputs: list[str]
    ) -> list[list[float]]:
        pass

    @abc.abstractmethod
    def connect(
        self,
        api_key: str,
    ) -> websockets.connect:
        pass



import rich
from httpx import HTTPStatusError

from zlipy.services.errors_handler.utils import pretty_time


class RequestErrorFormatter:
    def format(self, exc_value) -> str:
        if not isinstance(exc_value, HTTPStatusError):
            return f"{exc_value}"

        if exc_value.response is None:
            return f"{exc_value}"

        status_code = exc_value.response.status_code
        if status_code == 404:
            return "Cannot connect to the server. Please check your internet connection or try again later."
        elif status_code == 401:
            return "Unauthorized. Please check your API key."
        elif status_code == 403:
            return "Forbidden. You do not have permission to access this resource."
        elif status_code == 408:
            return "Request timeout. Please try again."
        elif status_code == 429:
            try:
                return f"Too many requests. Please slow down and try again in {pretty_time(int(exc_value.response.headers['Retry-After']))}."
            except (KeyError, ValueError):
                return "Too many requests. Please slow down and try again later."
        elif status_code == 503:
            return "Service unavailable. Please try again later."
        elif status_code >= 500 and status_code < 600:
            return "Server error. Please try again later."

        return f"{exc_value}"


class ErrorsHandler:
    def __init__(
        self,
        prefix: str | None = None,
        debug: bool = False,
    ) -> None:
        self.prefix = prefix
        self.debug = debug

        self.request_error_formatter = RequestErrorFormatter()

        self._handled_errors = False

    @property
    def handled_errors(self) -> bool:
        return self._handled_errors

    def __bool__(self) -> bool:
        return self.handled_errors

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            return False

        if isinstance(exc_value, HTTPStatusError):
            formatted_value = self.request_error_formatter.format(exc_value)
        else:
            formatted_value = f"{exc_value}"

        formatted_prefix = f"{self.prefix}: " if self.prefix else ""
        final_message = f"[red]{formatted_prefix}{formatted_value}[/red]"

        rich.print(final_message)

        if self.debug:
            raise exc_value

        self._handled_errors = True

        return True



from zlipy.services.api.clients import APIClient
from zlipy.services.api.interfaces import IAPIClient


class APIClientFactory:
    @staticmethod
    def create() -> IAPIClient:
        return APIClient()



from zlipy.config.constants import DEFAULT_CONFIG_FILENAME, DEFAULT_CONFIG_TEMPLATE


def init_config():
    """Initialize the configuration."""
    with open(DEFAULT_CONFIG_FILENAME, "w+") as f:
        f.write(DEFAULT_CONFIG_TEMPLATE)



import os

from zlipy.domain.filesfilter.interfaces import IFilesFilter, IProjectStructureLoader


class DefaultProjectStructureLoader(IProjectStructureLoader):
    def __init__(self, files_filter: IFilesFilter):
        self.files_filter = files_filter

    def _load(self, dir_path: str) -> list[str | dict]:
        result: list[str | dict] = []

        for item in os.listdir(dir_path):
            item_path = os.path.join(dir_path, item)

            if os.path.isdir(item_path):
                if sub_result := self._load(item_path):
                    result.append({item: sub_result})

            elif not self.files_filter.ignore(item_path):
                result.append(item)

        return result

    def load(self) -> list[str | dict]:
        return self._load(os.getcwd())



import asyncio

from rich.console import Console
from rich.progress import Progress


class LoadingAnimation:
    def __init__(self):
        self._stop_event = asyncio.Event()

    async def run(self):
        console = Console()
        progress = Progress(console=console, transient=True)
        task = progress.add_task("[cyan]Conducting analysis...", total=100)
        with Progress(transient=True) as progress:
            task = progress.add_task("Conducting analysis...", total=None)
            while not self._stop_event.is_set():
                progress.update(task, advance=0)
                await asyncio.sleep(0.0001)

        progress.stop()

    def stop(self):
        self._stop_event.set()

    async def start(self):
        self._stop_event.clear()
        await self.run()



import asyncio

from zlipy.config.interfaces import IConfig
from zlipy.services.client import ClientFactory
from zlipy.services.errors_handler import ErrorsHandler


def run(config: IConfig) -> None:
    with ErrorsHandler(
        prefix="Error during client initialization",
        debug=config.debug,
    ) as handler:
        client = ClientFactory.create(config=config)

    if handler.handled_errors:
        return

    asyncio.run(client.run())



from zlipy.domain.events.events import Event
from zlipy.domain.events.interfaces import IEvent


class EventFactory:
    @staticmethod
    def create(json_data: dict) -> IEvent:
        name = json_data.get("event")

        if name is None:
            raise ValueError("Event name is required")

        del json_data["event"]

        return Event(name, json_data)



from rich.console import Console

from zlipy.services.console.ublock_tty import UnblockTTY


async def aprint(*args, **kwargs):
    with UnblockTTY():
        console = Console()
        console.print(*args, **kwargs)


async def asimple_print(*args, **kwargs):
    with UnblockTTY():
        print(*args, **kwargs)



import configparser

from zlipy.config.configs import DefaultConfig
from zlipy.config.constants import DEFAULT_CONFIG_FILENAME
from zlipy.config.interfaces import IConfig


class ConfigFactory:
    @staticmethod
    def create(
        debug: bool = False,
        deep_dive: bool = False,
        disable_markdown_formatting: bool = False,
    ) -> IConfig:
        filename = DEFAULT_CONFIG_FILENAME
        config = configparser.ConfigParser()
        config.read(filename)

        if "settings" not in config.sections():
            raise ValueError(
                f"[bold red]settings[/] section not found in configuration file. Please, ensure you write it correctly inf your [bold red]{DEFAULT_CONFIG_FILENAME}[/] file"
            )

        if api_key := config["settings"].get("api_key"):
            ignored_patterns = []

            for pattern in config["settings"].get("ignored_patterns", "").split(","):
                if pattern_stripped := pattern.strip():
                    ignored_patterns.append(pattern_stripped)

            return DefaultConfig(
                api_key=api_key,
                debug=debug,
                deep_dive=deep_dive,
                disable_markdown_formatting=disable_markdown_formatting,
                ignored_patterns=ignored_patterns,
            )
        else:
            raise ValueError(
                "[bold red]api_key[/] not found in configuration file. Please, ensure you write it correctly inf your [bold red]{DEFAULT_CONFIG_FILENAME}[/] file"
            )



from zlipy.config.interfaces import IConfig


class DefaultConfig(IConfig):
    def __init__(
        self,
        api_key: str,
        debug: bool = False,
        deep_dive: bool = False,
        disable_markdown_formatting: bool = False,
        ignored_patterns: list[str] | None = None,
    ) -> None:
        super().__init__()
        self._api_key = api_key
        self._debug = debug
        self._deep_dive = deep_dive
        self._disable_markdown_formatting = disable_markdown_formatting
        self._ignored_patterns = ignored_patterns or []

    @property
    def api_key(self) -> str:
        return self._api_key

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def disable_markdown_formatting(self) -> bool:
        return self._disable_markdown_formatting

    @property
    def ignored_patterns(self) -> list[str]:
        return self._ignored_patterns

    @property
    def deep_dive(self) -> bool:
        return self._deep_dive



import asyncio
import json

import aioconsole  # type: ignore
import rich
import websockets
from rich.markdown import Markdown

from zlipy.config.interfaces import IConfig
from zlipy.domain.events import EventFactory, IEvent
from zlipy.domain.filesfilter import (
    FilesFilterFactory,
    FilesFilterTypes,
    ProjectStructureLoaderFactory,
)
from zlipy.domain.tools import ITool
from zlipy.services.api import IAPIClient
from zlipy.services.client.interfaces import IClient
from zlipy.services.console import aprint, asimple_print
from zlipy.services.console.loading_animation import LoadingAnimation


class Client(IClient):
    def __init__(
        self,
        api_client: IAPIClient,
        config: IConfig,
        tools: dict[str, ITool] | None = None,
    ) -> None:
        super().__init__()
        self.api_client = api_client
        self.config = config
        self.tools: dict[str, ITool] = tools or {}
        self.loading_animation = LoadingAnimation()

    async def _call_tool(self, tool_name: str, query: str):
        tool = self.tools.get(tool_name)

        return f"Tool {tool_name} not found" if tool is None else await tool.run(query)

    async def _pretty_print_message(self, message: str):
        if not self.config.disable_markdown_formatting:
            await aprint(Markdown(f"{message}"))
        else:
            await asimple_print(message)

    async def _debug_print(self, object):
        if self.config.debug:
            await aprint(object)

    async def _send_event(
        self, websocket: websockets.WebSocketClientProtocol, event: IEvent
    ):
        prepared_data = json.dumps({"event": event.name, **event.data})

        if self.config.debug:
            await self._debug_print(f"> Sending: {prepared_data}")

        await websocket.send(prepared_data)

    async def _handle_event(
        self, websocket: websockets.WebSocketClientProtocol, event: IEvent
    ):
        if event.name == "ToolCallEvent":
            if event.data["tool"] not in self.tools:
                await self._send_event(
                    websocket,
                    EventFactory.create(
                        {
                            "event": "ToolCallResponseEvent",
                            "error": f"Tool {event.data['tool']} not found",
                        }
                    ),
                )
                return
            else:
                if event.data["tool"] == "search":
                    await self._send_event(
                        websocket,
                        EventFactory.create(
                            {
                                "event": "SearchToolCallResponseEvent",
                                "documents": await self._call_tool(
                                    event.data["tool"], event.data["query"]
                                ),
                            }
                        ),
                    )
                elif event.data["tool"] == "load":
                    await self._send_event(
                        websocket,
                        EventFactory.create(
                            {
                                "event": "LoadFileToolCallResponseEvent",
                                "content": await self._call_tool(
                                    event.data["tool"], event.data["query"]
                                ),
                            }
                        ),
                    )

        if event.name == "WaitingForConfigurationEvent":
            await self._send_event(
                websocket,
                EventFactory.create(
                    {
                        "event": "ConfigurationEvent",
                        "tools": list(self.tools.keys()),
                        "project_structure": ProjectStructureLoaderFactory.create(
                            FilesFilterFactory.create(
                                FilesFilterTypes.DEFAULT,
                                self.config.ignored_patterns,
                            )
                        ).load(),
                        "boost": self.config.deep_dive,
                    }
                ),
            )

        if event.name == "AgentMessageEvent":
            await self._pretty_print_message(event.data["message"])

    def _stop_handle_events_condition(self, event: IEvent) -> bool:
        return event.name == "ReadyEvent"

    def _stop_animation_events_condition(self, event: IEvent) -> bool:
        return event.name in ["ReadyEvent", "AgentMessageEvent"]

    async def _handle_events(self, websocket: websockets.WebSocketClientProtocol):
        while True:
            response = json.loads(await asyncio.wait_for(websocket.recv(), timeout=300))

            await self._debug_print(f"< Received: {response}")

            event = EventFactory.create(response)

            if self._stop_animation_events_condition(event):
                self.loading_animation.stop()
                await asyncio.sleep(0.0005)  # Await for the animation to stop

            if self._stop_handle_events_condition(event):
                await self._debug_print("< Ready event received")
                break

            await self._handle_event(websocket, event)

    async def run(self):
        async with self.api_client.connect(api_key=self.config.api_key) as websocket:
            while websocket.open:
                try:
                    # Read until ready event is received with loading animation
                    # await self._handle_events(websocket)
                    await asyncio.gather(
                        self.loading_animation.start(), self._handle_events(websocket)
                    )

                    message = await aioconsole.ainput("Enter a message: ")

                    if not message:
                        await websocket.close()
                        await self._pretty_print_message("Connection closed")
                        break

                    await websocket.send(message)

                    await self._debug_print(f"> Sent: {message}")

                except Exception as e:
                    if isinstance(e, websockets.exceptions.ConnectionClosed):
                        await self._pretty_print_message("Connection closed")
                        break

                    import traceback

                    await aioconsole.aprint(traceback.format_exc())



from zlipy.domain.filesfilter.constants import GITIGNORE_FILENAME, FilesFilterTypes
from zlipy.domain.filesfilter.filters import (
    AllowedExtensionsFilesFilter,
    GitIgnoreFilesFilter,
    IgnoredFilesFilter,
    MergeFilesFilter,
)
from zlipy.domain.filesfilter.interfaces import IFilesFilter, IProjectStructureLoader
from zlipy.domain.filesfilter.loaders import DefaultProjectStructureLoader


class FilesFilterFactory:
    @staticmethod
    def create(
        files_filter_type: FilesFilterTypes = FilesFilterTypes.DEFAULT,
        ignore_patterns: list[str] | None = None,
    ) -> IFilesFilter:
        if files_filter_type == FilesFilterTypes.DEFAULT:
            if not ignore_patterns:
                ignore_patterns = [
                    "deeplake/",
                    "zlipy.ini",
                    "*.log",
                    "*.sql",
                    "*.json",
                    "*.csv",
                    "*.xml",
                    "*.txt",
                ]

            return MergeFilesFilter(
                GitIgnoreFilesFilter(GITIGNORE_FILENAME),
                AllowedExtensionsFilesFilter(),
                IgnoredFilesFilter(patterns=ignore_patterns),
            )

        if files_filter_type == FilesFilterTypes.GITIGNORE:
            return GitIgnoreFilesFilter(GITIGNORE_FILENAME)

        if files_filter_type == FilesFilterTypes.ALLOWED_EXTENSIONS:
            return AllowedExtensionsFilesFilter()

        raise ValueError(f"Unknown files filter type: {files_filter_type}")


class ProjectStructureLoaderFactory:
    @staticmethod
    def create(files_filter: IFilesFilter) -> IProjectStructureLoader:
        return DefaultProjectStructureLoader(files_filter)
